#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <ncurses.h>
#include <sys/time.h>
#include <time.h>
#include <sys/ioctl.h>
#include <math.h>
#include <signal.h>
#include <errno.h>
#include <libmpeg3.h>
#include <fftw3.h>

#include "error.h"
#include "snd_dev.h"
#include "fft.h"

#define DEFAULT_SND_DEV			"/dev/dsp"
#define DEFAULT_REFRESH_RATE		5
#define DEFAULT_MAX_N_PIXELS_PERCENTAGE	50.0
#define DEFAULT_MIN_SHOW_DURATION	2.0
#define DEFAULT_MAX_SHOW_DURATION	4.0

#define N_DRAW_TYPES 	4

#define min(x, y)  ((x) < (y) ? (x) : (y))
#define max(x, y)  ((x) > (y) ? (x) : (y))

int got_signal = 0;

void sighandler(int sig)
{
	got_signal = 1;
}

void determine_terminal_size(int *max_y, int *max_x)
{
	struct winsize size;

	*max_x = *max_y = 0;

	/* changed from 'STDIN_FILENO' as that is incorrect: we're
	 *          * outputting to stdout!
	 *                   */
	if (ioctl(1, TIOCGWINSZ, &size) == 0)
	{
		*max_y = size.ws_row;
		*max_x = size.ws_col;
	}

	if (!*max_x || !*max_y)
	{
		char *dummy = getenv("COLUMNS");
		if (dummy)
			*max_x = atoi(dummy);
		else
			*max_x = 80;

		dummy = getenv("LINES");
		if (dummy)
			*max_x = atoi(dummy);
		else
			*max_x = 24;
	}
}

int determine_number_of_pixels(float max_n_pixels_percentage, int max_y, int max_x)
{
	return ((float)max_y * max_x) * (max_n_pixels_percentage / 100.0);
}

double get_ts(void)
{
	struct timeval ts;
	struct timezone tz;

	if (gettimeofday(&ts, &tz) == -1)
		error_exit("gettimeofday() failed\n");

	return (((double)ts.tv_sec) + ((double)ts.tv_usec)/1000000.0);
}

void draw_method_0(int max_y, int max_x, short *left, short *right, int cur_n_samples, int draw_n_pixels)
{
	int loop, step;

	werase(stdscr);
	step = max(1, cur_n_samples / draw_n_pixels);
	for(loop=0; loop<cur_n_samples; loop += step)
	{
		int x = (max_x / 2) + (double)left [loop] * (max_x / 65536.0);
		int y = (max_y / 2) + (double)right[loop] * (max_y / 65536.0);
		int c = fabs(right[loop] - left[loop]);

		c %= 7;
		c++;
		if (x < 0) x = 0;
		if (y < 0) y = 0;
		if (x >= max_x) x = max_x - 1;
		if (y >= max_y) y = max_y - 1;

		wattron(stdscr, COLOR_PAIR(c));
		mvwprintw(stdscr, y, x, "*");
		wattroff(stdscr, COLOR_PAIR(c));
	}
}

void draw_method_1(int max_y, int max_x, short *left, short *right, int cur_n_samples, int draw_n_pixels)
{
	int step, loop;
	double scale_y = ((double)max_y) / 32768.0;
	double scale_x = ((double)max_x) / (double)cur_n_samples;

	werase(stdscr);

	step = cur_n_samples / max_x;

	for(loop=0; loop<cur_n_samples; loop += step)
	{
		int fill_y;
		int x = loop * scale_x;
		int half_y = max_y / 2;
		int left_y  = half_y - (fabs(left [loop]) * scale_y);
		int right_y = half_y + (fabs(right[loop]) * scale_y);

		if (left_y < 0) left_y = 0;
		if (right_y >= max_y) right_y = max_y - 1;

		wattron(stdscr, COLOR_PAIR(3));
		for(fill_y = half_y; fill_y > left_y; fill_y--) mvwprintw(stdscr, fill_y, x, "l");
		wattron(stdscr, A_BOLD);
		mvwprintw(stdscr, left_y,  x, "L");
		wattroff(stdscr, A_BOLD);
		wattroff(stdscr, COLOR_PAIR(3));
		wattron(stdscr, COLOR_PAIR(4));
		for(fill_y = half_y; fill_y < right_y; fill_y++) mvwprintw(stdscr, fill_y, x, "r");
		wattron(stdscr, A_BOLD);
		mvwprintw(stdscr, right_y, x, "R");
		wattroff(stdscr, A_BOLD);
		wattroff(stdscr, COLOR_PAIR(4));
	}
}

void draw_method_2(int max_y, int max_x, short *left, short *right, int cur_n_samples, int draw_n_pixels, fft_info *fft_data, double *buffer_out)
{
	int step, loop;
	int half_y = max_y / 2;
	double scale_y = 1.0;
	int half_sample_rate = cur_n_samples / 2;
	double scale_x = ((double)max_x) / (double)half_sample_rate;

	werase(stdscr);

	step = half_sample_rate / max_x;

	do_fft(fft_data, left, cur_n_samples, buffer_out);
	wattron(stdscr, COLOR_PAIR(5));
	for(loop=0; loop<half_sample_rate; loop += step)
	{
		int fill_y;
		int x = loop * scale_x;
		int left_y  = half_y - (fabs(buffer_out[loop]) * scale_y);

		if (left_y < 0) left_y = 0;

		for(fill_y = half_y; fill_y > left_y; fill_y--) mvwprintw(stdscr, fill_y, x, "l");
		wattron(stdscr, A_BOLD);
		mvwprintw(stdscr, left_y,  x, "L");
		wattroff(stdscr, A_BOLD);
	}
	wattroff(stdscr, COLOR_PAIR(5));

	do_fft(fft_data, right, cur_n_samples, buffer_out);
	wattron(stdscr, COLOR_PAIR(6));
	for(loop=0; loop<half_sample_rate; loop += step)
	{
		int fill_y;
		int x = loop * scale_x;
		int right_y  = half_y + (fabs(buffer_out[loop]) * scale_y);

		if (right_y >= max_y) right_y = max_y - 1;

		for(fill_y = half_y; fill_y < right_y; fill_y++) mvwprintw(stdscr, fill_y, x, "r");
		wattron(stdscr, A_BOLD);
		mvwprintw(stdscr, right_y, x, "R");
		wattroff(stdscr, A_BOLD);
	}
	wattroff(stdscr, COLOR_PAIR(6));
}

void draw_method_3(int max_y, int max_x, short *left, short *right, int cur_n_samples, int draw_n_pixels)
{
	int half_x = max_x / 2;
	int half_y = max_y / 2;
	double scale_y = (double)half_y / 32768.0;
	double end_y_l = (double)left [0] * scale_y;
	double end_y_r = (double)right[0] * scale_y;
	int x;

	werase(stdscr);

	wattron(stdscr, COLOR_PAIR(1));
	for(x=half_x; x>=0; x--)
		mvwprintw(stdscr, half_y + ((double)(half_x - x)*(end_y_l / (double)half_x)), x, "-");

	for(x=half_x; x<max_x; x++)
		mvwprintw(stdscr, half_y + ((double)(x - half_x)*(end_y_r / (double)half_x)), x, "-");
	wattroff(stdscr, COLOR_PAIR(1));
}

double get_rand(double max_value)
{
	return drand48() * max_value;
}

void version(void)
{
	printf("consolecandy v" VERSION ", (C) 2007 by folkert@vanheusden.com\n\n");
}

void help(void)
{
	version();

	printf("-f file      mp3 file to play\n");
	printf("-d device    sound-device to play through (default: %s)\n", DEFAULT_SND_DEV);
	printf("-i interval  refresh rate (default: %d)\n", DEFAULT_REFRESH_RATE);
	printf("-p perc      how many pixels to set (default percentage: %f)\n", DEFAULT_MAX_N_PIXELS_PERCENTAGE);
	printf("-m duration  minimum time a visualizer is shown (in seconds, default: %f)\n", DEFAULT_MIN_SHOW_DURATION);
	printf("-M duration  maximum time a visualizer is shown (in seconds, default: %f)\n", DEFAULT_MAX_SHOW_DURATION);
	printf("-V           show version and exit\n");
	printf("-h           this help\n");
}

int main(int argc, char *argv[])
{
	char *file = NULL;
	char *snddev = DEFAULT_SND_DEV;
	int refresh_rate = DEFAULT_REFRESH_RATE;
	float max_n_pixels_percentage = DEFAULT_MAX_N_PIXELS_PERCENTAGE;
	int slice_overflow_exit = 0;
	double min_show_duration = DEFAULT_MIN_SHOW_DURATION;
	double max_show_duration = DEFAULT_MAX_SHOW_DURATION;

	int sfd;
	int sample_rate = -1;
	int sample_size; /* sample_rate / refresh_rate */
	long number_of_samples_in_file;
	long n_slices = 0;
	short *left, *right;
	short *combined;
	double slice_duration;
	double draw_start_ts, draw_type_start_ts;
	int draw_n_pixels;
	int max_x = COLS, max_y = LINES;
	mpeg3_t *mf;
	int c;
	int draw_type = 0;
	double draw_type_show_length = 4.0;
	fft_info *fft_data;
	double *fft_buffer_out;
#ifdef M7
	int dummy;
#endif

	while((c = getopt(argc, argv, "f:d:i:p:m:M:Vh")) != -1)
	{
		switch(c)
		{
			case 'f':
				file = optarg;
				break;

			case 'd':
				snddev = optarg;
				break;

			case 'i':
				refresh_rate = atoi(optarg);
				if (refresh_rate < 1)
					error_exit("-i requires a value > 0\n");
				break;

			case 'p':
				max_n_pixels_percentage = atof(optarg);
				if (max_n_pixels_percentage < 0 || max_n_pixels_percentage > 100)
					error_exit("-p requires a value >=0 and <= 100\n");
				break;

			case 'm':
				min_show_duration = atof(optarg);
				if (min_show_duration <= 0.0)
					error_exit("-m requires a value >= 0\n");

			case 'M':
				max_show_duration = atof(optarg);
				if (max_show_duration <= min_show_duration)
					error_exit("-M requires a value >= the value of -m\n");

			case 'V':
				version();
				return 0;

			case 'h':
				help();
				return 0;

			default:
				help();
				return 1;
		}
	}

	if (!file)
	{
		help();
		error_exit("Please give an mp3-file to play.\n");
	}

#ifdef M7
	mf = mpeg3_open(file, &dummy);
#else
	mf = mpeg3_open(file);
#endif
	if (!mf)
		error_exit("Failed to open file %s\n", file);

	if (!mpeg3_has_audio(mf))
		error_exit("File has no audio stream(s).\n");

	if (mpeg3_total_astreams(mf) != 1)
		error_exit("File has %d streams while only 1 is supported.\n", mpeg3_total_astreams(mf));

	if (mpeg3_audio_channels(mf, 0) != 2)
		error_exit("Stream 0 has %d channels while 2 are required.\n", mpeg3_audio_channels(mf, 0));

	number_of_samples_in_file = mpeg3_audio_samples(mf, 0);

	sfd = open(snddev, O_RDWR);
	if (sfd == -1)
		error_exit("Failed opening audio device %s.\n", snddev);

	if (fcntl(sfd, F_SETFL, O_NONBLOCK) == -1)
		error_exit("Failure setting sound device to non-blocking.\n");

	set_soundcard_sample_format(sfd);
	set_soundcard_number_of_channels(sfd, mpeg3_audio_channels(mf, 0));
	sample_rate = mpeg3_sample_rate(mf, 0);
	set_soundcard_sample_rate(sfd, sample_rate);

	sample_size = sample_rate / refresh_rate;
	if (sample_size == 0)
		error_exit("Refresh rate (%d) is too high. Use at most %d.", refresh_rate, sample_rate);

	slice_duration = 1.0 / (double)refresh_rate;

	left  = malloc(sample_size * sizeof(short));
	right = malloc(sample_size * sizeof(short));
	combined = malloc(sample_size * 2 * sizeof(short));
	
	fft_data = alloc_fft(sample_rate);
	fft_buffer_out = malloc(sample_rate * sizeof(double)); /* only halve + 1 is needed in reality */

	if (!left || !right || !combined || !fft_data || !fft_buffer_out)
		error_exit("Memory allocation error.\n");

	initscr();
	start_color();

	init_pair(1, COLOR_WHITE,  COLOR_BLACK);
	init_pair(2, COLOR_YELLOW, COLOR_BLACK);
	init_pair(3, COLOR_RED,    COLOR_BLACK);
	init_pair(4, COLOR_GREEN,  COLOR_BLACK);
	init_pair(5, COLOR_BLUE,   COLOR_BLACK);
	init_pair(6, COLOR_MAGENTA, COLOR_BLACK);
	init_pair(7, COLOR_CYAN,   COLOR_BLACK);

	if (signal(SIGWINCH, sighandler) == SIG_ERR)
		error_exit("Failed to set signal handler for SIGWINCH.\n");

	determine_terminal_size(&max_y, &max_x);
	draw_n_pixels = determine_number_of_pixels(max_n_pixels_percentage, max_y, max_x);

	draw_type_start_ts = draw_start_ts = get_ts();
	for(;number_of_samples_in_file > 0;)
	{
		int loop, index = 0;
		char *send_pt = (char *)combined;
		int cur_n_samples = min(number_of_samples_in_file, sample_size);
		int bytes_to_send = cur_n_samples * 2 * 2;
		double drawing_took, drawing_should_have_taken;

		/* decode mp3 file */
		if (mpeg3_read_audio(mf, NULL, left, 0, cur_n_samples, 0))
			break;

		if (mpeg3_reread_audio(mf, NULL, right, 1, cur_n_samples, 0))
			break;

		number_of_samples_in_file -= cur_n_samples;
		n_slices++;

		/* did the terminal resize? */
		if (got_signal)
		{
			determine_terminal_size(&max_y, &max_x);
			draw_n_pixels = determine_number_of_pixels(max_n_pixels_percentage, max_y, max_x);

			if (ERR == resizeterm(max_y, max_x))
				error_exit("An error occured while resizing terminal(-window).\n");

			endwin();
			refresh();

			got_signal = 0;
		}

		/* play */
		for(loop=0; loop<cur_n_samples; loop++)
		{
			combined[index++] = left [loop];
			combined[index++] = right[loop];
		}
		while(bytes_to_send > 0)
		{
			int rc = write(sfd, send_pt, bytes_to_send);
			if (rc > 0)
			{
				send_pt += rc;
				bytes_to_send -= rc;
			}
			else
			{
				if (errno != EAGAIN && errno != EINTR)
					error_exit("write to sound-device failed\n");
			}
		}

		/* draw */
		if ((get_ts() - draw_type_start_ts) >= draw_type_show_length)
		{
			draw_type = lrand48() % N_DRAW_TYPES;
			draw_type_start_ts = get_ts();
			draw_type_show_length = min_show_duration + get_rand(max_show_duration - min_show_duration);
		}
		switch(draw_type)
		{
			case 0:
				draw_method_0(max_y, max_x, left, right, cur_n_samples, draw_n_pixels);
				break;

			case 1:
				draw_method_1(max_y, max_x, left, right, cur_n_samples, draw_n_pixels);
				break;

			case 2:
				draw_method_2(max_y, max_x, left, right, cur_n_samples, draw_n_pixels, fft_data, fft_buffer_out);
				break;

			case 3:
				draw_method_3(max_y, max_x, left, right, cur_n_samples, draw_n_pixels);
				break;
		}
		wrefresh(stdscr);
		doupdate();

		/* determine how long we need to sleep before we can draw the new frame */
		drawing_took = get_ts() - draw_start_ts;
		drawing_should_have_taken = slice_duration * (double)n_slices;
		if (drawing_took > drawing_should_have_taken && slice_overflow_exit)
			error_exit("Out of sync. Expected: %f, current: %f\n", drawing_should_have_taken, drawing_took);
	}

	endwin();

	close(sfd);
	mpeg3_close(mf);

	return 0;
}
